;
; Name: 	Protected mode micro-kernel
; Version:	0.90
; Date: 	Dec-16-1999
; Author:	Vladimir Arnost  <arnost@dcse.fee.vutbr.cz>
; Compiler:	TASM 3.2
;
; History:
;   0.00  Nov-22-1998	Project initiated
;   0.10  Nov-24-1998	Switching to PM and back to RM finally works
;   0.20  Nov-30-1998	Rewritten GDT/LDT allocation scheme
;   0.30  Dec-01-1998	Exception and IRQ handlers, Real-mode callbacks
;   0.40  Dec-03-1998	Text screen output routines, Exception reporting
;   0.50  Dec-08-1998	User code entry point callgate, Exception code and
;			stack segment dump
;   0.60  Dec-09-1998	User code entry point code fixed, API functions handler
;   0.70  Dec-10-1998	Source code reorganization, new API functions
;   0.72  Dec-14-1998	TSS segment changed from 16 to 32 bits, Fixed TSS32
;			STRUC declaration
;   0.80  Dec-15-1998	New function API_SETINT, Added IDT data alias descriptor
;			Screen size autodetection
;   0.81  Dec-17-1998	Minor changes
;   0.82  Jan-04-1999	Fixed a bug in get/set_cursor (x <-> y), Disabled
;			interrupts in get/set_cursor_offs, More cleanup code
;   0.90  Dec-16-1999	Exception handlers changed from 16 to 32-bit INTGATEs
;			Fixed improperly printed upper half of EFLAGS register
;			Fixed missing code and stack data in exception dumps
;			Limited the number of handled exceptions to EXC_MAX
;			Modified UpdateTable to recursively process all LDTs
;			 (it is no longer necessary to call it separately 
;			  to update all user code LDTs)
;			Modified SELECTOR macro to accept other LDT segments
;			Modified JMP/CALL_FAR macros for USE32 code segments
;			Removed unusable ret. addr from the stack in user code
;			Added API_PRINTDEC, API_SETSCRSIZE and API_DELAY
;
; WARNING: THIS CODE IS UNDER CONSTRUCTION. USE AT YOUR OWN RISK!
;

.386p
LOCALS
JUMPS
.CREF
DOSSEG
PAGE 60,132

KERN_VERSION	EQU	'0.90  Dec-16-1999'

RM_STACK_SIZE	EQU	128		; real-mode stack size (in words)
RMCB_STACK_SIZE EQU	256		; real-mode callback stack
PM_STACK_SIZE	EQU	256		; protected mode stack
EXC_MAX		EQU	100		; maximal number of handled exceptions

INCLUDE 	KERNEL.INC

OPENSEG		KERN_CODE,'CODE'
CLOSESEG	KERN_CODE
OPENSEG		KERN_DATA
CLOSESEG	KERN_DATA

dos_print	MACRO	msg
		mov	ah,9
  IFDIFI <msg>, <dx>
		mov	dx,offset msg
  ENDIF
		int	21h
ENDM

OPENSEG		KERN_CODE,'CODE'
		ASSUME	CS:KERN_CODE, DS:KERN_DATA, ES:KERN_DATA

;*** Program entry point
Kernel		PROC
Start:		mov	ax,SEG KERN_DATA
		mov	ds,ax
		mov	es,ax
		cld
;
		dos_print msg_start
;
		call	Check386
		mov	dx,offset msg_need386
		jc	@@error
		call	CheckV86
		mov	dx,offset msg_noV86
		jc	@@error
;
		call	RM_Init		; user-supplied initialization code
		call	detect_BIOS_screen
		call	calibrate_delay
;
		call	CreateGDT
		call	CreateIDT
;
		dos_print msg_enterPM
;
		call	Run
;
		call	sync_BIOS_cursor
		dos_print msg_leavePM
;
		call	RM_Cleanup	; user-supplied code
; exit to DOS
		mov	ax,4C00h
		int	21h
; print a message and exit
@@error:	dos_print dx
		mov	ax,4C01h
		int	21h
Kernel		ENDP

;*** Check for presence of 386 or a better processor
; In:		none
; Out:		AX=0,  CF=NC - 386 found
;		AX!=0, CF=CY - no 386
; Changes:	AH
Check386	PROC
		pushf
		mov	ax,7000h
		push	ax
		popf
		pushf
		pop	ax
		and	ax,07000h
		jz	@@1			; not found
		mov	ax,1			; found 386
@@1:		popf
		sub	ax,1
		ret
Check386	ENDP

;*** Check for V86 mode
; In:		none
; Out:		AX=0,  CF=NC - real mode
;		AX!=0, CF=CY - V86/protected mode
; Changes:	AX
CheckV86	PROC
		smsw	ax
		and	ax,1		; bit 0: Protected mode/V86 mode
		add	ax,-1
		inc	ax		; 0->0, NC  1->1, CY
		ret
CheckV86	ENDP

;*** Calculate linear address from real-mode Segment:Offset
; In:		_seg, _ofs
; Out:		DX:AX - 32-bit linear address
; Changes:	AX, DX
lin_addr	MACRO	_seg,_ofs
		mov	ax,_seg
		mov	dx,_seg
		shl	ax,4
		shr	dx,(16-4)
  IFDIFI <_ofs>,<0>
		add	ax,offset _ofs
		adc	dx,0
  ENDIF
ENDM

;*** Create initial Global and Local Descriptor Table entries
; In:		none
; Out:		none
; Changes:	AX, CX, DX, SI
CreateGDT	PROC
		push	ds
		mov	ax,SEG GRP_GDT
		mov	ds,ax
		xor	si,si
		mov	cx,offset GRP_GDT:GDT_end
		shr	cx,3
		call	UpdateTable
		pop	ds
		ret
CreateGDT	ENDP

;*** Update Descriptor Table entries, convert segments into linear addresses
; In:		DS:SI - table
;		CX    - number of entries
; Out:		none
; Changes:	AX, CX, DX, SI
UpdateTable	PROC	FAR
		jcxz	@@Ret
@@Loop:		mov	ax,WORD PTR [si].rights	; AL=rights, AH=ext
		test	al,10000000b
		jz	@@Skip			; skip Not Present segments
		test	ah,EXT_LINEAR
		jnz	@@Skip			; already linear address
		and	al,00011111b
		cmp	al,TYPE_LDT AND 00011111b
		jne	@@1
; this is a LDT entry, update it (a LDT may not contain another LDT entry)
		push	ds si cx
		mov	cx,[si].limit
		inc	cx
		shr	cx,3
		mov	ds,[si].baselo
		xor	si,si
		call	UpdateTable		; recursive call
		pop	cx si ds
; continue with normal entries
@@1:		mov	al,[si].rights
		and	al,00010100b
		cmp	al,00000100b		; skip Call/Int/Trap gates
		je	@@Skip
	 	lin_addr [si].baselo, 0 	; real-mode segment value
		mov	[si].baselo,ax		; bits 0:15 of linear addr
		mov	[si].basehi,dl		; bits 16:23
		mov	[si].basetop,dh 	; bits 24:31 (always zero)
		or	[si].ext,EXT_LINEAR	; mark is as processed
@@Skip:		add	si,8
		loop	@@Loop
@@Ret:		ret
UpdateTable	ENDP

;*** Create the Interrupt Descriptor Table
; In:		none
; Out:		none
; Changes:	AX
CreateIDT	PROC
; allow INT 1, INT 3 and INTO (INT 4) from any privilege-level (DPL3)
		push	ds
		mov	ax,SEG GRP_IDT
		mov	ds,ax
                mov	al,DPL3
                or	ds:[1*8].rights,al
                or	ds:[3*8].rights,al
                or	ds:[4*8].rights,al
		pop	ds
		ret
CreateIDT	ENDP

;*** Reprogram the master PIC (8259) IRQ vector base
; In:		AH - new IRQ base vector (normally 8), must be a multiple of 8
; Out:		none
; Changes:	AX
ReprogramPIC	PROC
		push	dx
		mov	dx,21h
		in	al,dx	; old IRQ mask
		push	ax
		mov	al,0FFh
		out	dx,al	; disable all IRQs temporarily
		call	@@delay
; magic follows...  :)
		mov	al,11h
		out	20h,al	; write ICW1 (Initialization Command Word 1)
		call	@@delay
		mov	al,ah
		out	dx,al	; ICW2: base interrupt vector (multiple of 8)
		call	@@delay
		mov	al,00000100b
		out	dx,al	; ICW3: slave 8259 is at int 2 (1 shl 2 = 4)
		call	@@delay
		mov	al,00000001b
		out	dx,al	; ICW4: 80x86 mode
		call	@@delay
; restore IRQ mask
		pop	ax
		out	dx,al
		pop	dx
		ret
@@delay:	in	al,80h	; insert a few wait states
		ret
ReprogramPIC	ENDP

;*** Set up protected mode environment, run user code and return back to RM
; In:		none
; Out:		AX=0, CF=NC  - OK
;		system crash - error
; Changes:	too much
Run		PROC
; save real-mode stack pointer and all registers
		pushfd
		pushad
		mov	save_RM_stack.OFFS,sp
		mov	save_RM_stack.SEGM,ss
		sgdt	RM_GDT_ptr
		sidt	RM_IDT_ptr
		cli
; disable NMI
		mov	al,8Dh	; select RTC status register D (read-only)
		out	70h,al
; reprogram master PIC to get rid of conflicting IRQs and exceptions
		mov	ah,68h	; IRQ0-7: INT 68h-6Fh, IRQ8-15: INT 70h-77h
		call	ReprogramPIC
; calculate GDT and IDT pointers
		lin_addr <SEG GRP_GDT>, 0	; GDT_table
		mov	GDT_base.LOWORD,ax
		mov	GDT_base.HIWORD,dx
		lin_addr <SEG GRP_IDT>, 0	; IDT_table
		mov	IDT_base.LOWORD,ax
		mov	IDT_base.HIWORD,dx
; load GDTR and IDTR
		lgdt	GDT_ptr
		lidt	IDT_ptr
; get ready to press the big red button...
		mov	eax,cr0
		or	al,1
		mov	cr0,eax
; load proper descriptors into all segment registers
		JMP_FAR S_KERN_PM_CODE,@@1
@@1:		mov	ax,S_KERN_PM_DATA
		mov	ds,ax
		mov	es,ax
		mov	fs,ax
		mov	gs,ax
		mov	ax,S_MAIN_STACK_0
		mov	ss,ax
		mov	esp,offset main_stack_0_top
		mov	ax,S_MAIN_TSS
		ltr	ax
		mov	ax,S_MAIN_LDT
		lldt	ax
                xor	eax,eax
                mov	cr2,eax		; clear last page-fault address
                mov	cr3,eax		; clear page directory address
		pushf
                pop	ax
                and	ax,not 4000h	; clear the NT (Nested Task) flag
                push	ax
                popf
                clts			; task not switched (yet)
		cld
		sti
; run user code
		mov	bx,PM_Entry.SEGM
                test	bl,0011b
                jz	@@Start		; enter ring 0 user code
; user code runs in Ring 1-3 (we must switch stacks)
		and	bx,0011b	; extract target CPL bits
		shl	bx,2
		push	DWORD PTR main_stacks[bx]
@@Start:	xor	eax,eax		; clean up
		xor	ebx,ebx
		xor	ecx,ecx
		xor	edx,edx
		xor	esi,esi
		xor	edi,edi
		xor	ebp,ebp
; CALL fails for transfers to rings != CPL(0)
; CALL GATE fails for transfers to lower priority RPL > CPL(0)
; FAR RET simulates a return from a CALL GATE (CPL(1..3) -> CPL(0) -> RET)
		push	PM_Entry
                retf			; go...

; prepare for a switch to real-mode
PM_Return:	cli
; set segment limits back to 64KB
		JMP_FAR S_KERN_RM_CODE,@@2
@@2:		mov	ax,S_KERN_RM_DATA
		mov	ds,ax
		mov	es,ax
		mov	fs,ax
		mov	gs,ax
		mov	ax,S_KERN_RM_STACK
		mov	ss,ax
		mov	sp,save_RM_stack.OFFS
;
		pushf
                pop	ax
                and	ax,not 4000h	; clear the NT (Nested Task) flag
                push	ax
                popf
                clts
; flip the switch...
		mov	eax,cr0
		and	al,not 1
		mov	cr0,eax
; jump to real-mode
		JMP_FAR <seg @@3>,@@3
; restore all registers
@@3:		mov	ax,SEG KERN_DATA
		mov	ds,ax
		mov	es,ax
		lss	sp,save_RM_stack
; restore default master PIC settings
		mov	ah,8h	; IRQ0-7: INT 08h-0Fh, IRQ8-15: INT 70h-77h
		call	ReprogramPIC
;
		lgdt	RM_GDT_ptr
		lidt	RM_IDT_ptr
		popad
		popfd
; enable NMI
		mov	al,0Dh
		out	70h,al
		xor	ax,ax			; OK
		ret
Run		ENDP

;*** Dummy protected mode user code entry point
; In:		none
; Out:		none
; Changes:	none
Dummy_PM_Main	PROC
		retf
Dummy_PM_Main	ENDP

;*** Switch temporarily into real-mode, run user code and return back to PM
; In:		DX:AX - seg:offs of real-mode code (far proc)
; Out:		?
; Changes:	?
Call_RM_Proc	PROC
		pushfd
		pushad
		push	ds es fs gs
		mov	bx,ax			; save target offset
		mov	ax,S_KERN_PM_DATA
		mov	ds,ax
		push	save_RMCB_stack.SEGM32
		push	save_RMCB_stack.OFFS32
		cli
		mov	save_RMCB_stack.OFFS32,esp
		mov	save_RMCB_stack.SEGM32,ss
; set segment limits back to 64KB
		JMP_FAR S_KERN_RM_CODE,@@0
@@0:		mov	ax,S_KERN_RM_DATA
		mov	ds,ax
		mov	es,ax
		mov	fs,ax
		mov	gs,ax
		mov	ax,S_KERN_RM_STACK
		mov	ss,ax
		mov	sp,0			; no ints/pushes allowed here!
; switch back to real-mode
		mov	eax,cr0
		and	al,not 1
		mov	cr0,eax
; jump to real-mode code segment
		JMP_FAR <seg @@1>,@@1
; restore all registers
@@1:		mov	ax,SEG KERN_DATA
		mov	ds,ax
		mov	es,ax
		mov	fs,ax
		mov	gs,ax
		mov	ax,SEG KERN_STACK_RMCB
		mov	ss,ax
		mov	sp,offset rmcb_stack_top
; restore default master PIC settings
		mov	ah,8h	; IRQ0-7: INT 08h-0Fh, IRQ8-15: INT 70h-77h
		call	ReprogramPIC
;
		lgdt	RM_GDT_ptr
		lidt	RM_IDT_ptr
; call user code
		pushf			; user code may return via IRET
		push	cs offset @@2	; return address
		push	dx bx
		retf			; jump to DX:BX
; return to protected mode
@@2:		cli
		mov	ax,SEG KERN_DATA
		mov	ds,ax
; reprogram master PIC to get rid of conflicting IRQs and exceptions
		mov	ah,68h	; IRQ0-7: INT 68h-6Fh, IRQ8-15: INT 70h-77h
		call	ReprogramPIC
; load GDTR and IDTR
		lgdt	GDT_ptr
		lidt	IDT_ptr
; enable protected mode
		mov	eax,cr0
		or	al,1
		mov	cr0,eax
; jump to protected mode code segment
		JMP_FAR S_KERN_PM_CODE,@@3
@@3:		mov	ax,S_KERN_PM_DATA
		mov	ds,ax
		lss	esp,PWORD PTR save_RMCB_stack
		pop	save_RMCB_stack.OFFS32
		pop	save_RMCB_stack.SEGM32
		pop	gs fs es ds
		popad
		popfd
		xor	ax,ax			; OK
@@Busy: 	ret
Call_RM_Proc	ENDP

;**********************************************************************

EVEN

;*** Exception entry points
IRP	i,<0,1,2,3,4,5,6,7,8,9,15,16,18,19,20,21,22,23,24,25,26,27,28,29,30,31>
exc_&i&_entry:	push	DWORD PTR 0	;; push a dummy error code
		push	i		;; exception number
		jmp	Exc_handler
ENDM
EVEN
IRP	i,<10,11,12,13,14,17>		;; error code pushed by the CPU
exc_&i&_entry:	push	i		;; exception number
		jmp	short Exc_handler
ENDM

; layout of stack on entry to the exception handler
  EXCSTK_CR3	EQU	-16
  EXCSTK_CR2	EQU	-12
  EXCSTK_CR0	EQU	-8
  EXCSTK_LDT	EQU	-4
  EXCSTK_TSS	EQU	-2
exception_stack	STRUC
  EXCSTK_EDI	dd	?		; BP+0
  EXCSTK_ESI	dd	?
  EXCSTK_EBP	dd	?
  EXCSTK_ESP	dd	?
  EXCSTK_EBX	dd	?
  EXCSTK_EDX	dd	?
  EXCSTK_ECX	dd	?
  EXCSTK_EAX	dd	?
  EXCSTK_SS	dw	?		; BP+20h
  EXCSTK_GS	dw	?
  EXCSTK_FS	dw	?
  EXCSTK_ES	dw	?
  EXCSTK_DS	dw	?
  EXCSTK_NUM	dw	?		; exception number
  EXCSTK_ERR	dd	?		; error code
  EXCSTK_EIP	dd	?		; return address
  EXCSTK_CS	dw	?, ?
  EXCSTK_EFLAGS	dd	?
  EXCSTK_ESPx	dd	?		; stack saved on privilege-level change
  EXCSTK_SSx	dw	?, ?
ENDS

;*** Protected mode exception handler
; In:		STACK[+2] - error code (DWORD)
;		STACK[+0] - exception number (WORD)
; Out:		none
; Changes:	none
Exc_handler	PROC
		push	ds es fs gs ss
		pushad
		mov	bp,sp
		sub	sp,16
		mov	ax,S_KERN_PM_DATA
		mov	ds,ax
		mov	es,ax
		mov	fs,ax
                mov	ax,ss
		mov	gs,ax
; check for cycled exceptions
		dec	exc_counter
		jz	@@TooMany
;
		str	[bp-2]		; task register
		sldt	[bp-4]		; LDT
		mov	eax,cr0
		mov	[bp-8],eax	; CR0
		mov	eax,cr2
		mov	[bp-12],eax	; CR2
		mov	eax,cr3
		mov	[bp-16],eax	; CR3
		test	[bp+EXCSTK_CS],0011b	; has privilege-level changed?
		jnz	@@1
		add	[bp+EXCSTK_ESP],20	; adjust ESP (it isn't restored
                jmp	short @@2		; by POPAD, so I can change it)
@@1:		mov	eax,[bp+EXCSTK_ESPx]	; copy SS:ESP saved by the CPU
		mov	[bp+EXCSTK_ESP],eax	; into the variable where we
		mov	ax,[bp+EXCSTK_SSx]	; expect it (just for printing)
		mov	[bp+EXCSTK_SS],ax
;
@@2:		call	Exc_print
;
		mov	bx,[bp+EXCSTK_NUM]
		test	exc_info[bx],EXC_CONT
		jz	@@Stop
; clean up
		mov	sp,bp
		popad
                add	sp,2		; do not restore SS (not changed)
		pop	gs fs es ds
		add	sp,6		; discard exc. number and error code
		db	66h
		iret			; 32-bit iret

@@TooMany:	mov	si,offset exc_msg_toomany
		call	print_form_str
@@Stop:		jmp	PM_Return	; shutdown
Exc_handler	ENDP

;*** Print an exception reason message
; In:		GS:BP - exception stack frame
; Out:		none
; Changes:	none
Exc_print	PROC
; exception number and name
		mov	si,offset exc_msg_exc_num
		call	print_form_str
		mov	bx,[bp+EXCSTK_NUM]
		shl	bx,1
		mov	si,exc_names[bx]
                call	print_str
; address
		mov	si,offset exc_msg_exc_at
                call	print_form_str
; error code
		mov	bx,[bp+EXCSTK_NUM]
		test	exc_info[bx],EXC_ERRCODE
                jz	@@1
		mov	si,offset exc_msg_errc
		call	print_form_str
; registers
@@1:		mov	si,offset exc_msg_reg
		call	print_form_str
; code dump
		mov	si,offset exc_msg_code
		call	print_str
		mov	esi,[bp+EXCSTK_EIP]
		mov	di,[bp+EXCSTK_CS]
		verr	di
		jnz	@@10			; ZF=NZ: unreadable segment
		mov	es,di
		sub	esi,8			; ES:ESI - code
		cmc
		sbb	ebx,ebx			; EBX=-1 if ESI < 0, else EBX=0
		and	esi,ebx			; 
		mov	bx,16			; BX - number of bytes to dump
		db	66h
		lsl	di,di			; EDI - segment limit
@@2:		cmp	esi,[bp+EXCSTK_EIP]
		jne	@@3
		mov	al,'*'			; star marks the exception EIP
		call	print_char
@@3:		cmp	esi,edi
		ja	@@5
		mov	dl,es:[esi]		; read 1 byte of code
		mov	cx,2
		call	print_hex
@@4:		mov	al,' '			; print space
		call	print_char
		inc	esi
		dec	bx
		jnz	@@2
		jmp	short @@10
@@5:		push	ax			; beyond segment limit
		mov	ax,'?' or 0700h		; print question marks
		call	print_char
		call	print_char
		pop	ax
		jmp	short @@4
; stack dump
@@10:		mov	si,offset exc_msg_stack
		call	print_str
		mov	esi,[bp+EXCSTK_ESP]
		mov	bx,[bp+EXCSTK_SS]
		verr	bx
		jnz	@@20			; ZF=NZ: unreadable segment
		mov	es,bx
		db	66h
		lsl	di,bx			; EDI - limit
		mov	bx,12
@@11:		cmp	esi,edi
		jae	@@13
		mov	dx,es:[esi]
		mov	cx,4
		call	print_hex
@@12:		mov	al,' '
		call	print_char
		add	esi,2
		dec	bx
		jnz	@@11
		jmp	short @@20
@@13:		push	ax
		mov	ax,'?' or 0700h
		call	print_char
		call	print_char
		call	print_char
		call	print_char
		pop	ax
		jmp	short @@12
;
@@20:		mov	si,offset exc_msg_crlf
		call	print_form_str
		ret
Exc_print	ENDP

;*** Print a formatted string
; In:		DS:SI - ASCIIZ string to print
;		AH - text attribute
;		GS:BP - frame for argument substitutions
; Out:		SI - at the end of string
; Changes:	SI, AL, CX, EDX
print_form_str	PROC
		push	di
		cld
@@Loop:		lodsb
		or	al,al
		jz	@@End
		cmp	al,'~'
		je	@@Attr
		cmp	al,'#'
		jne	@@Print
@@Number:	lodsb			; offset
		movsx	di,al
		mov	edx,gs:[bp+di]
                lodsb			; digits
                mov	cl,al
                xor	ch,ch
                call	print_hex
		jmp	short @@Loop
@@Attr:		lodsb
		mov	ah,al		; attribute
		jmp	short @@Loop
@@Print:	call	print_char
		jmp	short @@Loop
@@End:		pop	di
		ret
print_form_str	ENDP

;*** IRQ entry points
ALIGN 4
IRP	i,<0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>
irq_&i&_entry:	push	i		;; interrupt request number
		jmp	short IRQ_handler
ENDM

IRQ_NUM_OFFS	EQU	14h		; stack offset of IRQ number

;*** Protected mode IRQ handler
; In:		STACK - IRQ number
; Out:		none
; Changes:	none
IRQ_handler	PROC
		push	ds es
		pusha
		mov	ax,S_KERN_PM_DATA
		mov	ds,ax
		mov	es,ax
		mov	bp,sp
		mov	bx,[bp+IRQ_NUM_OFFS]	; IRQ number
; update exception counter
comment ~
		mov	ax,S_VGA_TEXT
		mov	es,ax
		mov	ax,bx
                DIGIT_TO_HEX
		mov	ah,70h
		push	bx
		shl	bx,2
		add	bx,(1*80+40)*2
		mov	WORD PTR es:[bx],ax
		inc	WORD PTR es:[bx+160]
		pop	bx
;~
		mov	ax,bx
		shl	bx,1
		call	IRQ_handlers[bx]
;
		popa
		pop	es ds
		add	sp,2		; discard IRQ number
		iret
IRQ_handler	ENDP

;*** Acknowledge low IRQ (0-7) to the master PIC
IRQ_low 	PROC
		mov	al,20h
		out	20h,al		; EOI
		ret
IRQ_low 	ENDP

;*** Acknowledge high IRQ (8-15) to the slave and master PICs
IRQ_high	PROC
		mov	al,20h
		out	0A0h,al
		out	20h,al		; EOI
		ret
IRQ_high	ENDP

;*** Pass control down to real-mode IRQ handler
; In:		AX - IRQ number
IRQ_pass_down	PROC
		mov	bx,ax
		mov	bl,IRQ_vectors[bx]
		shl	bx,2
		push	es
		mov	ax,S_SEG_0000
		mov	es,ax
		mov	ax,es:[bx].OFFS
		mov	dx,es:[bx].SEGM
		pop	es
		call	Call_RM_proc
		ret
IRQ_pass_down	ENDP

;*** IRQ1: Keyboard interrupt
IRQ_1		PROC
		in	al,60h		; read the key scan-code
;
		mov	al,20h
		out	20h,al		; EOI
		ret
IRQ_1		ENDP

;**********************************************************************

;*** Print a string at the console, update cursor position
; In:		DS:SI - ASCIIZ string to print
;		AH - text attribute
; Out:		SI - at the end of string
; Changes:	SI, AL
print_str	PROC
		cld
                jmp	short @@1
@@Loop:		call	print_char
@@1:		lodsb
		or	al,al
		jnz	@@Loop
		ret
print_str	ENDP

;*** Print a hexadecimal number
; In:		EDX - number
;		CX - number of digits (1-8)
;		AH - text attribute
; Out:		none
; Changes:	AL
print_hex	PROC
		jcxz	@@Ret
		push	cx
		ror	edx,cl
		ror	edx,cl
		ror	edx,cl
		ror	edx,cl
@@Loop:		rol	edx,4
		mov	al,dl
                and	al,0Fh
                DIGIT_TO_HEX
		call	print_char
		loop	@@Loop
		pop	cx
@@Ret:		ret
print_hex	ENDP

;*** Print an unsigned decimal number
; In:		EDX - number
;		CX - number of digits (0-10), 0-unformatted
;		AH - text attribute
; Out:		none
; Changes:	AL
print_dec	PROC
		push	si eax ebx cx edx
		mov	si,offset print_dec_buf[15]
                cmp	cx,15
                jbe	@@1
		xor	cx,cx
@@1:		mov	ebx,10		; radix
                mov	eax,edx
@@Loop:		xor	edx,edx
		div	ebx
                add	dl,'0'
                dec	si
                mov	[si],dl
                test	cx,cx
		jnz	@@2
                test	eax,eax		; if CX=0, then stop if EAX reaches 0
		jnz	@@Loop
		jmp	short @@3
@@2:		loop	@@Loop
@@3:		pop	edx cx ebx eax
		call	print_str
		pop	si
@@Ret:		ret
print_dec	ENDP

;*** Print a single character at the console, scroll if necessary and
;*** update cursor position. Handles basic TTY control codes: CR, LF, TAB, BS
; In:		AL - character to print
;		AH - attribute byte
; Out:		none
; Changes:	none
print_char	PROC
		push	es ax bx cx dx di
                call	get_cursor_offs	; BX - cursor offset
		mov	es,fs:VGA_selector
		mov	cx,fs:VGA_columns
                cmp	al,13
                je	@@CR
                cmp	al,10
                je	@@LF
                cmp	al,9
                je	@@TAB
                cmp	al,8
                je	@@BS
; print ordinary character
@@Print:	mov	di,bx
                mov	es:[bx+di],ax	; print it
                inc	bx		; advance the cursor
@@Test:		cmp	bx,fs:VGA_size
                jb	@@Set
@@Scroll:	sub	bx,cx
		mov	dx,1
                mov	al,' '
		call	scroll_screen
                jmp	short @@Set
; Carriage return, move to column 0
@@CR:		mov	ax,bx
		xor	dx,dx
		div	cx		; AX <- Y, DX <- X
                sub	bx,dx		; subtract the X position
                jmp	short @@Set
; Line feed
@@LF:		add	bx,cx
                jmp	short @@Test
; Tab stops are every 8 columns
@@TAB:		push	ax
		mov	ax,bx
		xor	dx,dx
		div	cx		; AX <- Y, DX <- X
                mov	ax,dx
                add	dx,8
                and	dx,not 7	; round down to a multiple of 8
                sub	dx,ax		; DX - number of spaces to print
                pop	ax
		mov	al,' '
		mov	di,bx
@@TAB1:		mov	es:[bx+di],ax	; print it
                inc	bx		; advance the cursor
                inc	di
		cmp	bx,fs:VGA_size
                jae	@@Scroll	; stop at column 0 and scroll
                dec	dx
                jnz	@@TAB1
                jmp	short @@Set
; Backspace
@@BS:		cmp	bx,1
		adc	bx,-1		; decrease BX if not zero
; Set actual cursor position
@@Set:		call	set_cursor_offs	; BX - new cursor offset
@@Ret:		pop	di dx cx bx ax es
		ret
print_char	ENDP

;*** Scroll the screen up DX lines
; In:		DX - number of lines to scroll
;		AL - character to fill empty space with
;		AH - attribute byte
;		ES - screen selector
; Out:		none
; Changes:	none
scroll_screen	PROC
		push	cx si di
		cld
                mov	cx,fs:VGA_rows
		mov	si,fs:VGA_size	; SI = VGA_columns * VGA_rows
		xor	di,di
                sub	cx,dx
                jbe	@@1		; DX >= CX: clear the screen
		imul	cx,fs:VGA_columns  ; CX = (VGA_rows - lines) * VGA_columns
                sub	si,cx
                shl	si,1
; scroll
		rep	movs WORD PTR es:[si], es:[di]
; fill the gap
                sub	si,di
                shr	si,1
@@1:		mov	cx,si
		rep	stosw
;
                pop	di si cx
		ret
scroll_screen	ENDP

;*** Clear the screen 
; In:		AH - attribute byte
; Out:		none
; Changes:	none
clear_screen	PROC
		push	es ax cx di
		cld
		mov	cx,fs:VGA_size	; SI = VGA_columns * VGA_rows
		mov	es,fs:VGA_selector
		xor	di,di
                mov	al,' '
		rep	stosw
;
		xor	bx,bx
                call	set_cursor_offs
                pop	di cx ax es
		ret
clear_screen	ENDP

;*** get cursor position
; In:		none
; Out:		AL - column (0..maxX-1)
;		AH - row (0..maxY-1)
;		BX - cursor offset (y * maxX + x)
; Changes:	BX
get_cursor	PROC
		call	get_cursor_offs
		mov	ax,bx
		div	BYTE PTR fs:VGA_columns
		xchg	al,ah
		ret
get_cursor	ENDP

;*** set cursor position
; In:		AL - column (0..maxX-1)
;		AH - row (0..maxY-1)
; Out:		BX - cursor offset (y * maxX + x)
; Changes:	BX
set_cursor	PROC
		push	ax
		mov	bl,al		; x
                xor	bh,bh
		mov	al,BYTE PTR fs:VGA_columns
		mul	ah		; y * maxX
		add	bx,ax
                call	set_cursor_offs
		pop	ax
		ret
set_cursor	ENDP

;*** read cursor position
; In:		none
; Out:		BX - cursor offset (y * columns + x)
; Changes:	none
get_cursor_offs	PROC
		push	ax dx
		pushf
		cli
		mov	dx,fs:VGA_CRTC_port ; VGA CRT controller register index
                mov	al,0Eh		; select cursor position (high byte)
                out	dx,al
                inc	dx
                in	al,dx		; VGA CRT controller data register
                mov	bh,al
                dec	dx
;
                mov	al,0Fh		; select cursor position (low byte)
                out	dx,al
		inc	dx
                in	al,dx
                mov	bl,al
		popf
                pop	dx ax
		ret
get_cursor_offs	ENDP

;*** write cursor position
; In:		BX - cursor offset
; Out:		none
; Changes:	none
set_cursor_offs	PROC
		push	ax dx
		pushf
		cli
		mov	dx,fs:VGA_CRTC_port ; VGA CRT controller register index
                mov	al,0Eh		; select cursor position (high byte)
                out	dx,al
                inc	dx
		mov	al,bh
                out	dx,al		; VGA CRT controller data register
                dec	dx
;
                mov	al,0Fh		; select cursor position (low byte)
                out	dx,al
                inc	dx
                mov	al,bl
                out	dx,al
		popf
                pop	dx ax
		ret
set_cursor_offs	ENDP

;*** read screen size from BIOS variables (real mode only!)
; In:		none
; Out:		none
; Changes:	AX, DX
detect_BIOS_screen PROC
		push	es
		mov	ax,40h
		mov	es,ax
		mov	al,es:[4Ah]	; 0040:004A - screen columns
		mov	VGA_columns,ax
		mov	dx,ax
		mov	al,es:[84h]	; 0040:0084 - screen rows - 1
		or	al,al
		jnz	@@1
		mov	al,24
@@1:		inc	al
		mov	VGA_rows,ax
		mul	dx
		mov	VGA_size,ax
		mov	ax,es:[63h]	; 0040:0063 - CRT controller port
		mov	VGA_CRTC_port,ax
		mov	al,es:[49h]	; 0040:0049 - screen mode
		mov	dx,0B800h
		cmp	al,7
		jne	@@2
		mov	dx,0B000h	; mono mode 7
@@2:		mov	ax,SEG GRP_GDT
		mov	es,ax
		ASSUME	ES:GRP_GDT
		mov	es:d_vga_text.baselo,dx	; baselo = real mode segment
		ASSUME	ES:NOTHING
		pop	es
		ret
detect_BIOS_screen ENDP

;*** synchronize BIOS cursor position (real mode only!)
; In:		none
; Out:		none
; Changes:	AX, BX, DX
sync_BIOS_cursor PROC
		call	get_cursor
		mov	dx,ax		; DH=row, DL=column
		mov	ah,2		; Fn 2: Set Cursor
		mov	bh,0		; video page
		int	10h
		ret
sync_BIOS_cursor ENDP

;*** calibrate delay loop (real mode only!)
; In:		none
; Out:		none
; Changes:	EAX, BX, ECX, EDX, ES
calibrate_delay	PROC
		xor	ax,ax
                mov	es,ax
                mov	di,046Ch	; 0000:046C - system timer (55ms/tick)
; wait for a new timer tick
		mov	dx,es:[di]
@@Wait:		cmp	dx,es:[di]
		je	@@Wait
; calibrate
		mov	eax,0FFFFFFFFh - (55/2)	; max delay minus rounding
                call	delay_loop
                not	eax		; make the value positive
; divide by 55 to get a number of iterations per 1ms
                xor	edx,edx
                mov	ecx,55		; 55ms/tick
                div	ecx
                mov	delay_1ms,eax
		ret
calibrate_delay	ENDP

;*** delay loop (real and protected mode)
; In:		EAX - maximal number of passes
;		ES:DI - address of byte to keep waiting until it changes
; Out:		none
; Changes:	EAX, DX
delay_loop	PROC
                mov	dx,es:[di]
@@Loop:		sub	eax,1
		jc	@@Ret
                cmp	dx,es:[di]
                je	@@Loop
@@Ret:		ret
delay_loop	ENDP

;**********************************************************************

; layout of stack in the API dispatcher
API_stack	STRUC
  APISTK_FS	dw	?		; BP+0
  APISTK_BP	dw	?
  APISTK_IP	dw	?
  APISTK_CS	dw	?
  APISTK_FUNC	dw	?		; API function number (arg 1)
ENDS

;*** API function entry point
; In:		function-specific
; Out:		function-specific
; Changes:	function-specific
API_Entry	PROC	FAR
		LOCAL	addr:WORD = @@size
		push	bp
                push	fs
                mov	bp,sp
                sub	sp,@@size
                push	bx
		mov	bx,S_KERN_PM_DATA
		mov	fs,bx
;
                mov	bx,[bp+APISTK_FUNC]
                cmp	bx,API_MAX
                jbe	@@1
                mov	bx,offset API_invalid
                jmp	short @@2
@@1:		shl	bx,1
                mov	bx,fs:API_handlers[bx]
@@2:		mov	addr,bx
                pop	bx
;
                call	addr
;
		mov	sp,bp
                pop	fs
                pop	bp
		ret	2
API_Entry	ENDP

;*** Invalid API Fn handler
; In:		none
; Out:		CF=CY
; Changes:	none
API_invalid	PROC
		pusha
		push	ds gs
		mov	ax,S_KERN_PM_DATA
		mov	ds,ax
		mov	si,offset api_msg_invalid
                mov	ax,ss
                mov	gs,ax
		call	print_form_str
		pop	gs ds
                popa
                stc			; error indicator
		ret
API_invalid	ENDP

;*** API Fn 0: Exit program
; In:		none
; Out:		does not return
; Changes:	none
API_EXIT_	PROC
		jmp	PM_Return
API_EXIT_	ENDP

;*** API Fn 1: Print a single character, interpret control codes (CR,LF,BS,HT)
; In:		AL - character
;		AH - text attribute
; Out:		none
; Changes:	none
API_PRINT_	PROC
		jmp	print_char
API_PRINT_	ENDP

;*** API Fn 2: Print a hexadecimal number
; In:		EDX - number
;		CX - number of digits (1-8)
;		AH - text attribute
; Out:		none
; Changes:	AL
API_PRINTHEX_	PROC
		jmp	print_hex
API_PRINTHEX_	ENDP

;*** API Fn 3: Print an unsigned decimal number
; In:		EDX - number
;		CX - number of digits (0-10), 0-unformatted
;		AH - text attribute
; Out:		none
; Changes:	AL
API_PRINTDEC_	PROC
		push	ds
		push	fs
		pop	ds		; DS=S_KERN_PM_DATA
		call	print_dec
		pop	ds
		ret
API_PRINTDEC_	ENDP

;*** API Fn 4: Print zero-terminated string
; In:		DS:SI - string address
;		AH - text attribute
; Out:		none
; Changes:	SI, AX
API_PRINTSTR_	PROC
                push	ax dx
		mov	ax,[bp+APISTK_CS]
                mov	dx,ds
                arpl	dx,ax		; adjust privilege to application CPL
                mov	ds,dx
                pop	dx ax
		call	print_str
		ret
API_PRINTSTR_	ENDP

;*** API Fn 5: Print zero-terminated formatted string (not C-like syntax!)
; In:		DS:SI - string address
;		ES:DI - arguments pointer
;		AH - text attribute
; Out:		none
; Changes:	SI, AX
API_PRINTFMT_	PROC
                push	cx edx bp ds gs
                push	ax dx
		mov	ax,[bp+APISTK_CS]
                mov	dx,ds
                arpl	dx,ax		; adjust privilege to application CPL
                mov	ds,dx		; adjusted DS:SI
                mov	dx,es
                arpl	dx,ax
                mov	gs,dx
		mov	bp,di		; GS:BP = adjusted ES:DI
                pop	dx ax
		call	print_form_str
                pop	gs ds bp edx cx
		ret
API_PRINTFMT_	ENDP

;*** API Fn 6: Clear screen
; In:		AH - text attribute
; Out:		none
; Changes:	none
API_CLRSCR_	PROC
		jmp	clear_screen
API_CLRSCR_	ENDP

;*** API Fn 7: Scroll screen
; In:		DX - number of lines to scroll
;		AH - attribute byte
; Out:		none
; Changes:	AL
API_SCROLL_	PROC
		push	es
		mov	es,fs:VGA_selector
                mov	al,' '
                call	scroll_screen
                pop	es
		ret
API_SCROLL_	ENDP

;*** API Fn 8: Set cursor position
; In:		AL - column (0..maxX-1)
;		AH - row (0..maxY-1)
; Out:		BX - cursor offset (y * maxX + x)
; Changes:	BX
API_SETCURSOR_	PROC
		jmp	set_cursor
API_SETCURSOR_	ENDP

;*** API Fn 9: Get cursor position
; In:		none
; Out:		AL - column (0..maxX-1)
;		AH - row (0..maxY-1)
;		BX - cursor offset (y * maxX + x)
; Changes:	BX
API_GETCURSOR_	PROC
		jmp	get_cursor
API_GETCURSOR_	ENDP

;*** API Fn 10: Get screen selector and size
; In:		none
; Out:		AX - columns
;		DX - rows
;		CX - columns*rows
;		BX - text mode screen selector
; Changes:	AX, BX, CX, DX
API_GETSCRSIZE_	PROC
		mov	ax,fs:VGA_columns
		mov	dx,fs:VGA_rows
		mov	cx,fs:VGA_size
		mov	bx,fs:VGA_selector
		ret
API_GETSCRSIZE_	ENDP

;*** API Fn 11: Set screen selector and size
; In:		AX - columns
;		DX - rows
;		BX - text mode screen selector
; Out:		CX - columns*rows
; Changes:	CX
API_SETSCRSIZE_	PROC
		push	ax dx
		mov	fs:VGA_columns,ax
		mov	fs:VGA_rows,dx
		mul	dx
		mov	fs:VGA_size,ax
		mov	fs:VGA_selector,bx
		mov	cx,ax
		pop	dx ax
		ret
API_SETSCRSIZE_	ENDP

;*** API Fn 12: Get key code from BIOS keyboard queue (does not wait)
;*** IRQ 1 _must_ be handled by the BIOS for this to work correctly
; In:		none
; Out:		AL - key code
;		AH - extended code
;		CF=CF - no key available
;		CF=NC - key code in AX
; Changes:	AX
API_GETKEY_	PROC
		push	ds bx
		mov	ax,S_SEG_0000
		mov	ds,ax
;
		mov	bx,ds:[41Ah]	; queue head
                cmp	bx,ds:[41Ch]	; queue tail
                je	@@NoKey
                mov	ax,ds:[400h][bx]	; buffer is between 41Eh...43Dh
                add	bx,2
                cmp	bx,3Ch
		jbe	@@1
                mov	bx,1Eh
@@1:		mov	ds:[41Ah],bx	; update queue head
		clc
		jmp	short @@Ret
;
@@NoKey:	xor	ax,ax
		stc
@@Ret:		pop	bx ds
		ret
API_GETKEY_	ENDP

;*** API Fn 13: Call real-mode function
; In:		DX:AX - seg:offs of real-mode code (far proc)
; Out:		AX = 0
; Changes:	?
API_CALLRM_	PROC
		jmp	Call_RM_Proc
API_CALLRM_	ENDP

;*** API Fn 14: Get & set interrupt descriptor table entry
; In:		AL - interrupt vector number
;		AH - 0: AL means interrupt/exception vector number
;		     1: AL means IRQ number (is automatically remapped to
;			interrupt vector number)
;		SI:EDI - selector:offset32 of the new interrupt handler
;		CX - new gate type, rights and params word
; Out:		SI:EDI - selector:offset32 of the previous interrupt handler
;		CX - previous gate type, rights and params word
; Changes:	BX
API_SETINT_	PROC
		push	ax
; calculate IRQ vector number
		test	ah,ah
                jz	@@2
                cmp	al,15
                cmc				; reverse the test
                jb	@@Ret			; don't allow IRQ > 15
  IF (IRQ_0_BASE+8) EQ (IRQ_8_BASE)
                add	al,IRQ_0_BASE
  ELSE
                cmp	al,8
                jae	@@1
                add	al,IRQ_0_BASE
                jmp	short @@2
@@1:		add	al,IRQ_8_BASE-8
  ENDIF
; calculate IDT entry address
@@2:		mov	bl,al
		xor	bh,bh
                shl	bx,3
                push	ds
                mov	ax,S_MAIN_IDT_DATA
                mov	ds,ax
		pushf
                cli				; just for sure
; swap data in the IDT
		xchg	ds:[bx].offslo,di
		xchg	ds:[bx].segmnt,si
                rol	edi,16
		xchg	WORD PTR ds:[bx].params,cx
		xchg	ds:[bx].offshi,di
                rol	edi,16
;
                popf
                pop	ds
@@Ret:		pop	ax
		ret
API_SETINT_	ENDP

;*** API Fn 15: Delay ECX milliseconds
; In:		ECX - delay interval in milliseconds (ms)
; Out:		none
; Changes:	none
API_DELAY_	PROC
		jecxz	@@Ret
		push	es di eax ecx dx
		mov	ax,S_SEG_0000
		mov	es,ax
		mov	di,0400h		; some constant value (es:[di])
@@Loop:		mov	eax,fs:delay_1ms
		call	delay_loop
		loopd	@@Loop
		pop	dx ecx eax di es
@@Ret:		ret
API_DELAY_	ENDP

kern_code_end	LABEL	WORD
CLOSESEG	KERN_CODE

;**********************************************************************

OPENSEG		KERN_DATA

;*** error messages
msg_start	db	'Protected Mode micro-Kernel  Version ', KERN_VERSION, CRLF, EOL
msg_need386	db	'Error: You must have at least a 386 processor', CRLF, EOL
msg_noV86	db	'Error: The processor is already in V86 mode', CRLF, EOL
msg_enterPM	db	'Entering protected mode...', CRLF, EOL
msg_leavePM	db	'Leaving protected mode...', CRLF, EOL

ALIGN 16

GDT_ptr 	LABEL	PWORD
GDT_limit	dw	offset GRP_GDT:GDT_end-1        ; GDT limit
GDT_base	dd	?			        ; linear base address
IDT_ptr 	LABEL	PWORD
IDT_limit	dw	offset GRP_IDT:IDT_end-1        ; IDT limit
IDT_base	dd	?			        ; linear base address

PM_Entry	LABEL	DWORD
		dw	offset Dummy_PM_Main,S_KERN_PM_CODE	; S_DUMMY_ENTRY

save_RM_stack	dd	?			; saved RM stack pointer
RM_GDT_ptr	dp	?
RM_IDT_ptr	dp	?

save_RMCB_stack dd	?			; saved RM callback stack
		dw	?

main_stacks	dw	offset main_stack_0_top,S_MAIN_STACK_0
		dw	offset main_stack_1_top,S_MAIN_STACK_1
		dw	offset main_stack_2_top,S_MAIN_STACK_2
		dw	offset main_stack_3_top,S_MAIN_STACK_3

; exception handlers

EXC_TRAP	EQU	00b		; ret. addr = next instruction
EXC_FAULT	EQU	01b		; ret. addr = current instruction
EXC_ABORT	EQU	10b		; ret. addr = invalid?
EXC_ERRCODE	EQU	0100b		; error code pushed on the stack
EXC_ERRPAGE	EQU	1100b		; error code is special (page fault)
EXC_CONT	EQU	00010000b	; allow continuation of execution

;EXC_ERRCODES	 EQU	 0100111110000000000b  ; which excs have error code (bit field))

exc_info	db	EXC_FAULT		; 0: divide by zero
		db	EXC_TRAP		; 1: debug exception (fault/trap)
		db	EXC_TRAP		; 2: NMI
		db	EXC_TRAP or EXC_CONT	; 3: breakpoint
		db	EXC_TRAP		; 4: overflow
		db	EXC_FAULT		; 5: BOUND range exceeded
		db	EXC_FAULT		; 6: invalid opcode
		db	EXC_FAULT		; 7: device not available
		db	EXC_ABORT		; 8: double fault
		db	EXC_ABORT		; 9: FPU segment overrun [obsolete]
		db	EXC_FAULT or EXC_ERRCODE ; 10: invalid TSS
		db	EXC_FAULT or EXC_ERRCODE ; 11: segment not present
		db	EXC_FAULT or EXC_ERRCODE ; 12: stack fault
		db	EXC_FAULT or EXC_ERRCODE ; 13: general protection fault
		db	EXC_FAULT or EXC_ERRPAGE ; 14: page fault
		db	EXC_ABORT		; 15: reserved
		db	EXC_FAULT		; 16: floating point error
		db	EXC_FAULT or EXC_ERRCODE ; 17: alignment check
		db	EXC_ABORT		; 18: machine check
		db	(31-18) dup (EXC_ABORT)	; 19-31: undefined

exc_counter	dw	EXC_MAX

exc_0_name	db	'Divide Error',0
exc_1_name	db	'Debugger Single Step',0
exc_2_name	db	'Non-Maskable Interrupt',0
exc_3_name	db	'Breakpoint',0
exc_4_name	db	'Overflow',0
exc_5_name	db	'BOUND Range Exceeded',0
exc_6_name	db	'Invalid Opcode',0
exc_7_name	db	'Device Not Available',0
exc_8_name	db	'Double Fault',0
exc_9_name	db	'Coprocessor Segment Overrun',0
exc_10_name	db	'Invalid TSS',0
exc_11_name	db	'Segment Not Present',0
exc_12_name	db	'Stack Fault',0
exc_13_name	db	'General Protection Fault',0
exc_14_name	db	'Page Fault',0
exc_16_name	db	'Floating-Point Error',0
exc_17_name	db	'Alignment Check',0
exc_18_name	db	'Machine-Check',0
exc_undef_name	db	'Undefined Exception',0

exc_msg_toomany	db	'~',0Ch,CRLF,'Too many exceptions. Program aborted.','~',07h,CRLF,0
exc_msg_exc_num	db	'~',0Eh,CRLF,'Exception #',EXCSTK_NUM,2,': ',0
exc_msg_exc_at	db	' at #',EXCSTK_CS,4,':#',EXCSTK_EIP,8,0
exc_msg_errc	db	', error code #',EXCSTK_ERR,8,0
exc_msg_reg	db	'~',0Fh,CRLF,'EAX=#',EXCSTK_EAX,8,'  EBX=#',EXCSTK_EBX,8
		db	'  ECX=#',EXCSTK_ECX,8,'  EDX=#',EXCSTK_EDX,8
		db	'  EFLAGS=#',EXCSTK_EFLAGS,8
		db	CRLF,'ESI=#',EXCSTK_ESI,8,'  EDI=#',EXCSTK_EDI,8
		db	'  EBP=#',EXCSTK_EBP,8,'  ESP=#',EXCSTK_ESP,8
		db	CRLF,' CS=#',EXCSTK_CS,4,'  DS=#',EXCSTK_DS,4
		db	'  ES=#',EXCSTK_ES,4,'  FS=#',EXCSTK_FS,4
		db	'  GS=#',EXCSTK_GS,4,'  SS=#',EXCSTK_SS,4
		db	'  TSS=#',EXCSTK_TSS,4,'  LDT=#',EXCSTK_LDT,4
		db	CRLF,'CR0=#',EXCSTK_CR0,8,'  CR2=#',EXCSTK_CR2,8
		db	'  CR3=#',EXCSTK_CR3,8,CRLF,0
exc_msg_code	db	'Code:   ',0
exc_msg_stack	db	CRLF,'Stack: *',0
exc_msg_crlf	db	'~',07h,CRLF,0

EVEN

exc_names	dw	exc_0_name, exc_1_name, exc_2_name, exc_3_name
		dw	exc_4_name, exc_5_name, exc_6_name, exc_7_name
                dw	exc_8_name, exc_9_name, exc_10_name, exc_11_name
                dw	exc_12_name, exc_13_name, exc_14_name, exc_undef_name
                dw	exc_16_name, exc_17_name, exc_18_name, exc_undef_name
                dw	(31-19) dup (exc_undef_name)

; IRQ 0-15 handlers
IRQ_handlers	dw	IRQ_pass_down, IRQ_pass_down, IRQ_pass_down, IRQ_pass_down
		dw	IRQ_pass_down, IRQ_pass_down, IRQ_pass_down, IRQ_pass_down
		dw	IRQ_pass_down, IRQ_pass_down, IRQ_pass_down, IRQ_pass_down
		dw	IRQ_pass_down, IRQ_pass_down, IRQ_pass_down, IRQ_pass_down
;		dw	IRQ_high, IRQ_high, IRQ_high, IRQ_high
;		dw	IRQ_high, IRQ_high, IRQ_high, IRQ_high
IRQ_vectors	db	08h,09h,0Ah,0Bh,0Ch,0Dh,0Eh,0Fh
		db	70h,71h,72h,73h,74h,75h,76h,77h

; VGA output routines data
VGA_selector	dw	S_VGA_TEXT	; selector or real-mode segment (B800h)
VGA_CRTC_port	dw	03D4h		; VGA CRT controller register (3B4h in mono)
VGA_columns	dw	80
VGA_rows	dw	25
VGA_size	dw	80*25

print_dec_buf	db	16 dup (0)	; data buffer for print_dec function

delay_1ms	dd	0		; 1ms delay time coefficient

; API functions handlers
API_MAX		EQU	API_DELAY
API_handlers	dw	API_EXIT_, API_PRINT_, API_PRINTHEX_, API_PRINTDEC_
		dw	API_PRINTSTR_, API_PRINTFMT_, API_CLRSCR_, API_SCROLL_
		dw	API_SETCURSOR_, API_GETCURSOR_, API_GETSCRSIZE_, API_SETSCRSIZE_
		dw	API_GETKEY_, API_CALLRM_, API_SETINT_, API_DELAY_

api_msg_invalid	db	'~',0Fh,CRLF,'Undefined API function #',APISTK_FUNC,4
		db	' called from #',APISTK_CS,4,':#',APISTK_IP,4,'~',07h,CRLF,0

kern_data_end	LABEL  WORD
CLOSESEG	KERN_DATA

;**********************************************************************

OPENSEG		MAIN_GDT_START
GDT_table	LABEL	QWORD
d_null		descr	<>		; mandatory 0th descriptor
d_gdt_data	descr	<offset GRP_GDT:GDT_end-1,  seg GRP_GDT,	0,TYPE_DATA,0,?>
CLOSESEG	MAIN_GDT_START

OPENSEG		MAIN_GDT
d_kern_pm_code	descr	<offset kern_code_end-1,    seg KERN_CODE,	0,TYPE_CODE,0,?>
d_kern_pm_data	descr	<offset kern_data_end-1,    seg KERN_DATA,	0,TYPE_DATA,0,?>
d_main_stack_0	descr	<offset main_stack_0_top-1, seg MAIN_STACK_0,	0,TYPE_DATA or DPL0,0,?>
d_main_stack_1	descr	<offset main_stack_1_top-1, seg MAIN_STACK_1,	0,TYPE_DATA or DPL1,0,?>
d_main_stack_2	descr	<offset main_stack_2_top-1, seg MAIN_STACK_2,	0,TYPE_DATA or DPL2,0,?>
d_main_stack_3	descr	<offset main_stack_3_top-1, seg MAIN_STACK_3,	0,TYPE_DATA or DPL3,0,?>
d_kern_rm_code	descr	<0FFFFh,		    seg KERN_CODE,	0,TYPE_CODE,0,?>
d_kern_rm_data	descr	<0FFFFh,		    seg KERN_DATA,	0,TYPE_DATA,0,?>
d_kern_rm_stack descr	<0FFFFh,		    seg KERN_STACK_RM,	0,TYPE_DATA,0,?>
d_main_idt_data descr	<offset GRP_IDT:IDT_end-1,  seg GRP_IDT,	0,TYPE_DATA,0,?>
d_main_ldt	descr	<offset GRP_LDT:LDT_end-1,  seg GRP_LDT,	0,TYPE_LDT,0,?>
d_main_ldt_data descr	<offset GRP_LDT:LDT_end-1,  seg GRP_LDT,	0,TYPE_DATA,0,?>
d_main_tss	descr	<offset GRP_TSS:TSS_end-1,  seg MAIN_TSS,	0,TYPE_TSS32,0,?>
d_main_tss_data descr	<offset GRP_TSS:TSS_end-1,  seg MAIN_TSS,	0,TYPE_DATA,0,?>
d_vga_text	descr	<07FFFh,		    0B800h,		0,TYPE_DATA or DPL3,0,?>
d_vga_gfx	descr	<0FFFFh,		    0A000h,		0,TYPE_DATA or DPL3,0,?>
d_seg_0000	descr	<0FFFFh,		    00000h,		0,TYPE_DATA or DPL3,0,?>
d_api_gate	gatedescr <offset API_Entry,        S_KERN_PM_CODE,     1,TYPE_CALLG16 or DPL3,0>
CLOSESEG	MAIN_GDT

OPENSEG		MAIN_GDT_END
GDT_end 	LABEL	QWORD
CLOSESEG	MAIN_GDT_END

;**********************************************************************

OPENSEG		MAIN_LDT
LDT_table	LABEL	QWORD
		descr	<>		; make LDT segment size > 0
CLOSESEG	MAIN_LDT

OPENSEG		MAIN_LDT_END
LDT_end 	LABEL	QWORD
CLOSESEG	MAIN_LDT_END

;**********************************************************************

OPENSEG		MAIN_IDT
IDT_table	LABEL	QWORD
	IRP	i,<0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31>
		gatedescr <offset exc_&i&_entry,S_KERN_PM_CODE,0,TYPE_INTG32,0>
	ENDM
		ORG	IRQ_0_BASE*8
	IRP	i,<0,1,2,3,4,5,6,7>
		gatedescr <offset irq_&i&_entry,S_KERN_PM_CODE,0,TYPE_INTG16,0>
	ENDM
		ORG	IRQ_8_BASE*8
	IRP	i,<8,9,10,11,12,13,14,15>
		gatedescr <offset irq_&i&_entry,S_KERN_PM_CODE,0,TYPE_INTG16,0>
	ENDM
		ORG	80h*8
CLOSESEG	MAIN_IDT

OPENSEG		MAIN_IDT_END
IDT_end 	LABEL	QWORD
CLOSESEG	MAIN_IDT_END

;**********************************************************************

;*** Kernel Main Task State Segment (32-bit)
; only the stack pointers must be initialized, because this is an active task;
; the other values will be dumped here by the processor during a task switch;
; inactive tasks must be initialized completely, ALL tasks must have a valid
; ring 0 stack, otherwise no API calls nor exception handling will work (and
; the CPU would be instantly reset with no apparent reason) so take care!
OPENSEG		MAIN_TSS
;;		tss32	<>
	 	dw	0, 0			; tss32_link
		dd	offset main_stack_0_top	; tss32_stack0 ESP
		dw	S_MAIN_STACK_0, 0	; SS
		dd	offset main_stack_1_top	; tss32_stack1 ESP
		dw	S_MAIN_STACK_1, 0	; SS
		dd	offset main_stack_2_top	; tss32_stack2 ESP
		dw	S_MAIN_STACK_2, 0	; SS
		dd	0			; tss32_cr3
		dd	?			; tss32_eip
		dd	?			; tss32_eflags
		dd	?			; tss32_eax
		dd	?			; tss32_ecx
		dd	?			; tss32_edx
		dd	?			; tss32_ebx
		dd	?			; tss32_esp
		dd	?			; tss32_ebp
		dd	?			; tss32_esi
		dd	?			; tss32_edi
	 	dw	?, 0			; tss32_es
	 	dw	?, 0			; tss32_cs
	 	dw	?, 0			; tss32_ss
	 	dw	?, 0			; tss32_ds
	 	dw	?, 0			; tss32_fs
	 	dw	?, 0			; tss32_gs
		dw	S_MAIN_LDT, 0		; tss32_ldt
		dw	0			; tss32_t
		dw	offset GRP_TSS:TSS_end	; tss32_iomapbase
CLOSESEG	MAIN_TSS

OPENSEG		MAIN_TSS_END
TSS_end		LABEL	WORD
		db	0FFh	; end of I/O permission map
CLOSESEG	MAIN_TSS_END

;**********************************************************************

;*** real-mode stack
OPENSEG		KERN_STACK_RM,<STACK 'STACK'>
rm_stack_bottom dw	RM_STACK_SIZE dup (?)
rm_stack_top	LABEL	WORD
CLOSESEG	KERN_STACK_RM

;*** real-mode call-back stack
OPENSEG		KERN_STACK_RMCB,'STACK'
rmcb_stack_bottom dw	RMCB_STACK_SIZE dup (?)
rmcb_stack_top	LABEL	WORD
CLOSESEG	KERN_STACK_RMCB

;*** protected-mode stacks (main task, rings 0-3)
DECLARE_STACK	MAIN_STACK_0,PM_STACK_SIZE
DECLARE_STACK	MAIN_STACK_1,PM_STACK_SIZE
DECLARE_STACK	MAIN_STACK_2,PM_STACK_SIZE
DECLARE_STACK	MAIN_STACK_3,PM_STACK_SIZE

		END	Start

; End of file
