;
; Name: 	Protected mode micro-kernel macros and definitions
; Version:	0.90
; Date: 	Dec-16-1999
; Author:	Vladimir Arnost  <arnost@dcse.fee.vutbr.cz>
; Compiler:	TASM 3.2
;

;*** Useful constants
CR		EQU	0Dh
LF		EQU	0Ah
CRLF		EQU	CR,LF
EOL		EQU	'$'
NUL		EQU	0

SEGM		EQU	(WORD PTR 2)
OFFS		EQU	(WORD PTR 0)

SEGM32		EQU	(WORD PTR 4)
OFFS32		EQU	(DWORD PTR 0)

HIWORD		EQU	(WORD PTR 2)
LOWORD		EQU	(WORD PTR 0)

HIBYTE		EQU	(BYTE PTR 1)
LOBYTE		EQU	(BYTE PTR 0)

;**********************************************************************

;*** Descriptor table (GDT/LDT) entry
descr		STRUC
		limit	dw	?
		baselo	dw	?
		basehi	db	?
		rights	db	?	; TYPE_xxx
		ext	db	?	; EXT_xxx
		basetop db	?
ENDS

;*** Call/Task/Interrupt/Trap Gate entry
gatedescr	STRUC
		offslo	dw	?
		segmnt	dw	?
		params	db	?
		rights	db	?	; TYPE_xxx
		offshi	dw	?
ENDS

; segment descriptor types
TYPE_CODE	EQU	10011010b	; 9Ah - Code/Execute/Read/Not accessed
TYPE_DATA	EQU	10010010b	; 92h - Data/Read/Write/Not accessed
TYPE_CONST	EQU	10010000b	; 90h - Data/Read-only/Not accessed
TYPE_STACK	EQU	10010110b	; 96h - Data/Read/Write/Expand down/Not accessed

TYPE_TSS16	EQU	10000001b	; 81h - 16-bit TSS (Available)
TYPE_LDT	EQU	10000010b	; 82h - LDT
TYPE_TSS16B	EQU	10000011b	; 83h - 16-bit TSS (Busy)
TYPE_CALLG16	EQU	10000100b	; 84h - 16-bit Call Gate
TYPE_TASKG	EQU	10000101b	; 85h - Task Gate
TYPE_INTG16	EQU	10000110b	; 86h - 16-bit Interrupt Gate
TYPE_TRAPG16	EQU	10000111b	; 87h - 16-bit Trap Gate

TYPE_TSS32	EQU	10001001b	; 89h - 32-bit TSS (Available)
TYPE_TSS32B	EQU	10001011b	; 8Bh - 32-bit TSS (Busy)
TYPE_CALLG32	EQU	10001100b	; 8Ch - 32-bit Call Gate
TYPE_INTG32	EQU	10001110b	; 8Eh - 32-bit Interrupt Gate
TYPE_TRAPG32	EQU	10001111b	; 8Fh - 32-bit Trap Gate

; descriptor privilege levels (usage: TYPE_CODE or DPL3)
DPL0		EQU	00000000b	; DPL 0 (Descriptor Privilege Level)
DPL1		EQU	00100000b	; DPL 1
DPL2		EQU	01000000b	; DPL 2
DPL3		EQU	01100000b	; DPL 3

; descriptor extended attributes
EXT_LINEAR	EQU	00010000b	; do not translate real-mode segment
					; into linear address (it already is)
EXT_USE16	EQU	00000000b	; USE16 (16-bit code segment, default)
EXT_USE32	EQU	01000000b	; USE32 (32-bit code segment)
EXT_GRAN	EQU	10000000b	; granularity (limit is in 4K pages)

;**********************************************************************

;*** Selector flags (parameters to SELECTOR macro)
SEL_RPL0	EQU	00000000b	; RPL 0 (Requested Privilege Level)
SEL_RPL1	EQU	00000001b	; RPL 1
SEL_RPL2	EQU	00000010b	; RPL 2
SEL_RPL3	EQU	00000011b	; RPL 3
SEL_GDT 	EQU	00000000b	; Global descriptor
SEL_LDT 	EQU	00000100b	; Local descriptor

;*** Create selector value from descriptor entry offset
SELECTOR	MACRO	name,entry,table,rpl,type
  IFNB <type>
		GLOBAL	entry:type
  ELSE
		GLOBAL	entry:DESCR
  ENDIF
  IFIDNI <table>,<SEL_GDT>
      &name&	EQU	OFFSET GRP_GDT:entry + SEL_GDT + rpl
  ELSE
    IFIDNI <table>,<GRP_GDT>
      &name&	EQU	OFFSET GRP_GDT:entry + SEL_GDT + rpl
    ELSE
      IFIDNI <table>,<SEL_LDT>
      &name&	EQU	OFFSET GRP_LDT:entry + SEL_LDT + rpl
      ELSE
      ; other tables may be specified and they must be LDTs (GDT is only one)
      &name&	EQU	OFFSET table:entry + SEL_LDT + rpl
      ENDIF
    ENDIF
  ENDIF
ENDM

;*** Kernel Module Segment Selectors
SELECTOR	S_GDT_DATA,	d_gdt_data,	SEL_GDT, SEL_RPL0
SELECTOR	S_KERN_PM_CODE, d_kern_pm_code,	SEL_GDT, SEL_RPL0
SELECTOR	S_KERN_PM_DATA,	d_kern_pm_data,	SEL_GDT, SEL_RPL0
SELECTOR	S_MAIN_STACK_0,	d_main_stack_0,	SEL_GDT, SEL_RPL0
SELECTOR	S_MAIN_STACK_1,	d_main_stack_1,	SEL_GDT, SEL_RPL1
SELECTOR	S_MAIN_STACK_2,	d_main_stack_2,	SEL_GDT, SEL_RPL2
SELECTOR	S_MAIN_STACK_3,	d_main_stack_3,	SEL_GDT, SEL_RPL3
SELECTOR	S_KERN_RM_CODE,	d_kern_rm_code,	SEL_GDT, SEL_RPL0
SELECTOR	S_KERN_RM_DATA,	d_kern_rm_data,	SEL_GDT, SEL_RPL0
SELECTOR	S_KERN_RM_STACK,d_kern_rm_stack,SEL_GDT, SEL_RPL0
SELECTOR	S_SEG_0000,	d_seg_0000,	SEL_GDT, SEL_RPL3
SELECTOR	S_MAIN_IDT_DATA,d_main_idt_data,SEL_GDT, SEL_RPL0
SELECTOR	S_MAIN_LDT,	d_main_ldt,	SEL_GDT, SEL_RPL0
SELECTOR	S_MAIN_LDT_DATA,d_main_ldt_data,SEL_GDT, SEL_RPL0
SELECTOR	S_MAIN_TSS,	d_main_tss,	SEL_GDT, SEL_RPL0
SELECTOR	S_MAIN_TSS_DATA,d_main_tss_data,SEL_GDT, SEL_RPL0
SELECTOR	S_VGA_TEXT,	d_vga_text,	SEL_GDT, SEL_RPL3
SELECTOR	S_VGA_GFX,	d_vga_gfx,	SEL_GDT, SEL_RPL3
SELECTOR	S_API_CALL,	d_api_gate,	SEL_GDT, SEL_RPL3, GATEDESCR

;**********************************************************************

;*** Macros for creating paragraph-aligned segments
OPENSEG		MACRO	name,class,use
  IFDIFI <class>,<>
    &name&	SEGMENT USE16 PARA PUBLIC class use
  ELSE
    &name&	SEGMENT USE16 PARA PUBLIC 'DATA' use
  ENDIF
ENDM
CLOSESEG	MACRO	name
    &name&	ENDS
ENDM

;*** Descriptor Table Segments
GROUP GRP_GDT	MAIN_GDT_START,MAIN_GDT,MAIN_GDT_END

OPENSEG		MAIN_GDT_START
CLOSESEG	MAIN_GDT_START
OPENSEG		MAIN_GDT
CLOSESEG	MAIN_GDT
OPENSEG		MAIN_GDT_END
CLOSESEG	MAIN_GDT_END

GROUP GRP_LDT	MAIN_LDT,MAIN_LDT_END

OPENSEG		MAIN_LDT
CLOSESEG	MAIN_LDT
OPENSEG		MAIN_LDT_END
CLOSESEG	MAIN_LDT_END

GROUP GRP_IDT	MAIN_IDT,MAIN_IDT_END

OPENSEG		MAIN_IDT
CLOSESEG	MAIN_IDT
OPENSEG		MAIN_IDT_END
CLOSESEG	MAIN_IDT_END

IRQ_0_BASE	EQU	68h
IRQ_8_BASE	EQU	70h

GROUP GRP_TSS	MAIN_TSS,MAIN_TSS_END

OPENSEG		MAIN_TSS
CLOSESEG	MAIN_TSS
OPENSEG		MAIN_TSS_END
CLOSESEG	MAIN_TSS_END

;*** 16-bit Task State Segment (TSS)
tss16	STRUC
	tss16_link 	dw	?
	tss16_stack0	dw	?, ?		; SP, SS
	tss16_stack1	dw	?, ?		; SP, SS
	tss16_stack2	dw	?, ?		; SP, SS
	tss16_ip 	dw	?
	tss16_flags	dw	?
	tss16_ax 	dw	?		; registers
	tss16_cx 	dw	?
	tss16_dx 	dw	?
	tss16_bx 	dw	?
	tss16_sp 	dw	?
	tss16_bp 	dw	?
	tss16_si 	dw	?
	tss16_di 	dw	?
	tss16_es 	dw	?		; selectors
	tss16_cs 	dw	?
	tss16_ss 	dw	?
	tss16_ds 	dw	?
	tss16_ldt	dw	?		; task LDT
ENDS

;*** 32-bit Task State Segment (TSS)
tss32	STRUC
	tss32_link 	dw	?, 0
	tss32_stack0	dd	?		; ESP
			dw	?, 0		; SS
	tss32_stack1	dd	?
			dw	?, 0
	tss32_stack2	dd	?
			dw	?, 0
	tss32_cr3	dd	?
	tss32_eip	dd	?
	tss32_eflags	dd	?
	tss32_eax	dd	?		; registers
	tss32_ecx	dd	?
	tss32_edx	dd	?
	tss32_ebx	dd	?
	tss32_esp	dd	?
	tss32_ebp	dd	?
	tss32_esi	dd	?
	tss32_edi	dd	?
	tss32_es 	dw	?, 0		; selectors
	tss32_cs 	dw	?, 0
	tss32_ss 	dw	?, 0
	tss32_ds 	dw	?, 0
	tss32_fs 	dw	?, 0
	tss32_gs 	dw	?, 0
	tss32_ldt	dw	?, 0		; task LDT
	tss32_t		dw	0		; debug trap
	tss32_iomapbase	dw	?		; I/O map base address
ENDS

;*** Useful macros
JMP_FAR 	MACRO	_seg,_ofs
		db	0EAh	; JMP FAR PTR opcode
  IF @WordSize EQ 2
		dw	_ofs	; 16-bit code
  ELSE
		dd	_ofs	; 32-bit code
  ENDIF
		dw	_seg
ENDM

CALL_FAR	MACRO	_seg,_ofs
		db	09Ah	; CALL FAR PTR opcode
  IF @WordSize EQ 2
		dw	_ofs
  ELSE
		dd	_ofs
  ENDIF
		dw	_seg
ENDM

; Convert one digit (0-F) in AL into ASCII printable character '0'-'9','A'-'F'
DIGIT_TO_HEX	MACRO
		cmp	al,10
		sbb	al,69h
		das
ENDM

; Declare stack segment
DECLARE_STACK	MACRO	name,size
  OPENSEG	name,'STACK'
  &name&_bottom	LABEL	WORD
		dw	size dup (?)
  &name&_top	LABEL	WORD
  CLOSESEG	name
ENDM

;**********************************************************************

;*** Macro to call KERNEL API functions
CALL_API	MACRO	_func
		push	WORD PTR _func
		CALL_FAR S_API_CALL,0
ENDM

;*** API functions
API_EXIT	EQU	0
API_PRINT	EQU	1
API_PRINTHEX	EQU	2
API_PRINTDEC	EQU	3
API_PRINTSTR	EQU	4
API_PRINTFMT	EQU	5
API_CLRSCR	EQU	6
API_SCROLL	EQU	7
API_SETCURSOR	EQU	8
API_GETCURSOR	EQU	9
API_GETSCRSIZE	EQU	10
API_SETSCRSIZE	EQU	11
API_GETKEY	EQU	12
API_CALLRM	EQU	13
API_SETINT	EQU	14
API_DELAY	EQU	15

;**********************************************************************

;*** Global symbols accessible from user module
GLOBAL	UpdateTable:FAR
GLOBAL	RM_Init:FAR, RM_Cleanup:FAR
GLOBAL	PM_Entry:DWORD

; End of file
